﻿//---------------------------------------------------------------------------
//
//   PolygonCounter
//
//          Copyright(C) 2014, Tark
//
//  ポリゴン数をリアルタイムにカウント
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <memory>
#include <sstream>

#include "PolygonCounterWindow.h"
#include "PolygonCounter.h"
#include "MQSetting.h"

static PolygonCounter plugin;

// Constructor
// コンストラクタ
PolygonCounter::PolygonCounter() : m_Window(nullptr), m_bActivate(false),
	m_bUpdate(false), m_bAll(false), m_bVisible(true),
	m_bFreeze(false), m_bTriangle(true), m_bPolygon(true)
{
	setlocale(LC_CTYPE, "");
}

// Destructor
// デストラクタ
PolygonCounter::~PolygonCounter()
{
	if(m_Window != NULL){
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  GetPlugInID
//---------------------------------------------------------------------------
void PolygonCounter::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0xE3F7A830;
	*ID      = 0xBD9A1873;
}

//---------------------------------------------------------------------------
//  GetPlugInName
//---------------------------------------------------------------------------
const char *PolygonCounter::GetPlugInName(void)
{
	return "PolygonCounter Ver.1.1   Copyright(C) 2014, Tark.";
}

//---------------------------------------------------------------------------
//  EnumString
//---------------------------------------------------------------------------
const char *PolygonCounter::EnumString(void)
{
	return "PolygonCounter";
}

//---------------------------------------------------------------------------
//  Initialize
//---------------------------------------------------------------------------
BOOL PolygonCounter::Initialize()
{
	int posX = -1, posY = -1;
	{
		MQSetting *setting = OpenSetting();
		setting->Load("PolygonCounter_activate", m_bActivate, false);
		setting->Load("update", m_bUpdate, false);
		setting->Load("all", m_bAll, false);
		setting->Load("visible", m_bVisible, true);
		setting->Load("freeze", m_bFreeze, false);
		setting->Load("triangle", m_bTriangle, true);
		setting->Load("polygon", m_bPolygon, true);
		setting->Load("PosX", posX, -1);
		setting->Load("PosY", posY, -1);
		CloseSetting(setting);
	}
	if (m_Window == NULL){
		MQWindowBase mainwnd = MQWindow::GetMainWindow();
		m_Window = new PolygonCounterWindow(this, mainwnd);
		if(posX >= 0 && posY >= 0) {
			m_Window->SetPosX(posX);
			m_Window->SetPosY(posY);
		}
	}
	Activate(nullptr, m_bActivate);
	return TRUE;
}

//---------------------------------------------------------------------------
//  Exit
//    アプリケーションの終了
//---------------------------------------------------------------------------
void PolygonCounter::Exit()
{
	{
		MQSetting *setting = OpenSetting();
		setting->Save("PolygonCounter_activate", m_bActivate);
		setting->Save("update", m_bUpdate);
		setting->Save("all", m_bAll);
		setting->Save("visible", m_bVisible);
		setting->Save("freeze", m_bFreeze);
		setting->Save("triangle", m_bTriangle);
		setting->Save("polygon", m_bPolygon);
		setting->Save("PosX", m_Window->GetPosX());
		setting->Save("PosY", m_Window->GetPosY());
		CloseSetting(setting);
	}
	if (m_Window != NULL){
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  Activate
//---------------------------------------------------------------------------
BOOL PolygonCounter::Activate(MQDocument doc, BOOL flag)
{
//	m_doc = doc;
	if (m_Window == NULL) {
		return FALSE;
	}
	m_bActivate = flag ? true : false;
	m_Window->SetVisible(m_bActivate);

	return m_bActivate;
}

//---------------------------------------------------------------------------
//  IsActivated
//---------------------------------------------------------------------------
BOOL PolygonCounter::IsActivated(MQDocument doc)
{
	if (m_Window == NULL) {
		return FALSE;
	}
	count(doc);
	return m_bActivate;
}

//---------------------------------------------------------------------------
//  ExecuteCallback
///---------------------------------------------------------------------------
bool PolygonCounter::ExecuteCallback(MQDocument doc, void *option)
{
	CallbackInfo *info = (CallbackInfo*)option;
	return ((*this).*info->proc)(doc);
}

// コールバックの呼び出し
void PolygonCounter::Execute(ExecuteCallbackProc proc)
{
	CallbackInfo info;
	info.proc = proc;
	BeginCallback(&info);
}

// プラグインのベースクラスを返す
MQBasePlugin *GetPluginClass()
{
	return &plugin;
}

void PolygonCounter::OnNewDocument(MQDocument doc_, const char *filename, NEW_DOCUMENT_PARAM& param)
{
	count(doc_);
}

void PolygonCounter::OnObjectModified(MQDocument doc_)
{
	count(doc_);
}

void PolygonCounter::OnUpdateObjectList(MQDocument doc_)
{
	count(doc_);
}

BOOL PolygonCounter::OnUndo(MQDocument doc_, int undo_state)
{
	count(doc_);
	return FALSE;
}

BOOL PolygonCounter::OnRedo(MQDocument doc_, int redo_state)
{
	count(doc_);
	return FALSE;
}

void PolygonCounter::countobj(cmq::CMQObject obj, int &countv, int &countf, int &count2, int &count3, int &count4, int &count5)
{
	bool fleeze = false;
	if(m_bFreeze && (obj.getMirrorType() != MQOBJECT_MIRROR_NONE || obj.getPatchType() != MQOBJECT_PATCH_NONE || obj.getLatheType() != MQOBJECT_LATHE_NONE)) {
		// 鏡面、曲面、回転体指定されているオブジェクトはクローンを作ってフリーズしたものについてカウントする
		fleeze = true;
		obj = obj.clone();
		obj.freeze(MQOBJECT_FREEZE_ALL);
	}
	cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(&obj);
	for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
		countv++;
	}
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		int poly = fit->getFacePointCount();
		if(m_bTriangle) {
			// 三角形換算する
			switch(poly) {
			case 3: count3++; countf++; break;
			case 4: count4++; countf += 2; break;
			case 2: count2++; countf++; break;
			case 1: break;
			default: count5++; countf += poly - 2; break;
			}
		}
		else {
			// 三角形換算しない
			switch(poly) {
			case 3: count3++; countf++; break;
			case 4: count4++; countf++; break;
			case 2: count2++; countf++; break;
			case 1: break;
			default: count5++; countf++; break;
			}
		}
	}
	if(fleeze) {
		obj.deleteThis();
	}
}

void PolygonCounter::count(MQDocument doc_) {
	if(!m_bActivate || !m_bUpdate) {
		return;
	}
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject currentobj(doc.getCurrentObject());
	int ocount = doc.getObjectCount();
	int sumvc = 0, sumfc = 0, sum2c = 0, sum3c = 0, sum4c = 0, sum5c = 0;
	if(m_bAll) {
		// 全オブジェクト
		int sumva = 0, sumfa = 0, sum2a = 0, sum3a = 0, sum4a = 0, sum5a = 0;
		for(int i = 0; i < ocount; i++) {
			cmq::CMQObject obj_ = const_cast<cmq::CMQDocument&>(doc).getObject(i);
			if(obj_) {
				int countv = 0, countf = 0, count2 = 0, count3 = 0, count4 = 0, count5 = 0;
				if(!m_bVisible || obj_.getVisible() || obj_.m_obj == currentobj.m_obj) {
					// 可視オブジェクトのみカウント
					// ただしカレントオブジェクトは可視フラグにかかわらず対象とする
					countobj(obj_, countv, countf, count2, count3, count4, count5);
				}
				if(obj_.m_obj == currentobj.m_obj) {
					// カレントオブジェクト
					sumvc = countv;
					sumfc = countf;
					sum2c = count2;
					sum3c = count3;
					sum4c = count4;
					sum5c = count5;
				}
				if(!m_bVisible || obj_.getVisible()) {
					sumva += countv;
					sumfa += countf;
					sum2a += count2;
					sum3a += count3;
					sum4a += count4;
					sum5a += count5;
				}
			}
		}
		m_Window->setall(sumva, sumfa, sum2a, sum3a, sum4a, sum5a);
	}
	else {
		// カレントオブジェクトのみ
		countobj(currentobj, sumvc, sumfc, sum2c, sum3c, sum4c, sum5c);
	}
	m_Window->setcurrent(sumvc, sumfc, sum2c, sum3c, sum4c, sum5c);
}
